import operator
from datetime import datetime

from series.get import ReleaseMonitor, ReleaseHandler
from series.get.search import TorrentSearch
from series.get.model.link import Torrent
from series.handler import Job
from series.condition import StrictCondition
from series.get.handler import R
from series.util import now_unix

from golgi import configurable

from tek_utils.sharehoster.errors import ShareHosterError

from amino import List, Map, __


class Downgrade(Job):
    pass


class CheckCaching(Job):
    pass


class HasStandardDef(R):

    def __init__(self) -> None:
        super().__init__('resolutions', target='')

    @property
    def _desc(self):
        return 'has sd'

    @property
    def _oper(self):
        return operator.contains

    def _repr(self, item, match: bool) -> str:
        return 'sd' if match else 'no sd'


class Thresh(R):

    def __init__(self, attr) -> None:
        return super().__init__(attr, target=None)

    def _target(self, item):
        return item.downgrade_after_hours

    def _value(self, item):
        return datetime.now() - super()._value(item)

    @property
    def _oper(self):
        return operator.gt

    def _oper_repr(self, match):
        return '>' if match else '>'

    def _target_repr(self, item):
        return '{}h'.format(int(self._target(item).total_seconds() / 3600))

    @property
    def _desc(self):
        return '{} old'.format(super()._desc)


class AttrNull(R):

    def __init__(self, attr) -> None:
        super().__init__(attr, target=0)

    def _value(self, item):
        return super()._value(item) or 0

    @property
    def _desc(self):
        return '{} null'.format(super()._desc)


class AirdateNull(AttrNull):

    def __init__(self) -> None:
        super().__init__('airdate')

    def _value(self, item):
        return item.release.airdate


class AirdateThresh(Thresh):

    def __init__(self) -> None:
        super().__init__('airdate')

    def _value(self, item):
        return datetime.now() - item.release.airdate


class LastUpdateNull(AttrNull):

    def __init__(self) -> None:
        super().__init__('last_torrent_update_stamp')


@configurable(torrent=['search_engine'], get=['min_seeders'])
class TorrentCleaner(ReleaseHandler):

    def __init__(self, releases, *a, **kw):
        super().__init__(releases, 30, 'torrent cleaner', cooldown=3600, **kw)

    @property
    def _candidates(self):
        down = (
            self._no_cached_torrents_q
            .filter(ReleaseMonitor.downgrade_after > 0)
            .all()
        )
        caching = (
            self._no_cached_torrents_q
            .join(Torrent)
            .filter(Torrent.caching)
            .all()
        )
        return ((List.wrap(down) / Downgrade) +
                (List.wrap(caching) / CheckCaching))

    @property
    def _job_conditions(self):
        return Map({
            Downgrade: self._downgrade_conditions,
            CheckCaching: self._check_caching_conditions,
        })

    @property
    def _downgrade_conditions(self):
        return (
            ~AttrNull('downgrade_after') &
            ~HasStandardDef() &
            (
                (~LastUpdateNull() & Thresh('last_torrent_update')) |
                (LastUpdateNull() & ~AirdateNull() & AirdateThresh())
            )
        )

    @property
    def _check_caching_conditions(self):
        return StrictCondition(True)

    def _handle_job(self, job):
        monitor = job.item
        if isinstance(job, Downgrade):
            if not self._alternative_hd_releases(monitor):
                self._downgrade(monitor)
            self._reset(monitor)
        elif isinstance(job, CheckCaching):
            return self._check_caching(monitor)

    def _downgrade(self, monitor):
        self.log.info('Downgrading resolution of {}'.format(monitor.release))
        self._update(
            monitor,
            _resolutions=','.join(monitor.resolutions.cat('')),
        )

    def _reset(self, monitor):
        self.log.info('Resetting torrent {}'.format(monitor.release))
        self._releases.reset_torrent(monitor.id)
        self._update(
            monitor,
            last_torrent_update_stamp=now_unix()
        )

    def _alternative_hd_releases(self, monitor):
        search = TorrentSearch(monitor, self._description, self._search_engine,
                               self._min_seeders)
        search.start()
        search.join()
        # cannot be __, nesting breaks things
        g = lambda r: r.choose(monitor).is_just
        return search.result.exists(__.exists(g)).value

    def _check_caching(self, monitor):
        self._check_pending(monitor)

    def _check_pending(self, monitor):
        msg = 'Checking status of caching torrents for {}'
        self.log.debug(msg.format(monitor.release))
        for link in monitor.torrent_links:
            torrent = link.torrent
            try:
                if not link.cached and torrent.cached:
                    self._releases.torrent_cached(link.id)
                    msg = 'Flagging torrent as cached: {}'
                    self.log.info(msg.format(monitor.release))
                elif not link.dead and not link.cached and not torrent.caching:
                    self._releases.update_link(link.id, caching=False)
                    msg = 'Torrent not reporting as caching anymore: {}'
                    self.log.warning(msg.format(link.name))
                    self.log.info('status was: {}; message: {}'.format(
                        torrent.cacher.status, torrent.cacher.status_message))
            except ShareHosterError as e:
                msg = 'Error checking torrent status for {}: {}'
                self.log.error(msg.format(monitor.release, e))


__all__ = ('TorrentCleaner',)



import threading
from functools import wraps

import requests
from requests.exceptions import RequestException

from flask import Flask, jsonify, request

from golgi import configurable

from series.logging import Logging


def json_response(func):
    @wraps(func)
    def wrapper(*a, **kw):
        response = func(*a, **kw)
        if isinstance(response, tuple):
            rest = response[1:]
            response = response[0]
        elif response is None:
            rest = ()
            response = {}
        else:
            rest = ()
        return (jsonify(dict(response=response)),) + rest
    return wrapper


def route_decorator():
    routes = {}

    def _route(path, **kwargs):
        def add_route(func):
            configs = routes.setdefault(func.__name__, [])
            configs.append((path, kwargs,))
            return func
        return add_route
    return routes, _route


@configurable(general=['debug'])
class RestApi(threading.Thread, Logging):

    def __init__(self, name):
        threading.Thread.__init__(self)
        self.app = Flask(__name__)
        if self._debug:
            self.app.config['TESTING'] = True

    def run(self):
        self.log.info('Starting REST API.')
        self.setup_routes()
        try:
            self.app.run(host=self._rest_api_host, port=self._rest_api_port)
        except OSError as e:
            self.log.error('''Couldn't start REST API: {}'''.format(e))

    def _shutdown(self):
        func = request.environ.get('werkzeug.server.shutdown')
        func()
        return 'server shut down.'

    def stop(self):
        self.app.add_url_rule('/shutdown', 'shutdown', self._shutdown,
                              methods=['PUT'])
        uri = 'http://localhost:{}/shutdown'.format(self._rest_api_port)
        try:
            requests.put(uri)
        except RequestException as e:
            self.log.debug('api shutdown request failed: {}'.format(e))

    def setup_routes(self):
        for name, configs in self.routes.items():
            for index, config in enumerate(configs):
                path, kwargs = config
                func = getattr(self, name)
                self.app.add_url_rule(path, name + str(index),
                                      json_response(func), **kwargs)

__all__ = ['RestApi', 'route_decorator']



from golgi import configurable

from amino import __, List

from series.get.releases_facade import ReleasesFacade
from series.get.db import FileDatabase
from series.app import App
from series.get.shows_facade import ShowsFacade
from series.handler import Handler
from series.get.model.release import ReleaseMonitor
from series.get.handler import ReleaseHandler, ShowHandler
from series.get.model.show import Show
from series.get.util.format import (format_explain_release,
                                    format_explain_show, format_status)


@configurable(get=['db_path', 'run', 'omit', 'auto_upgrade_db'])
class SeriesGetD(App):
    _components = ['feed_poller', 'downloader', 'archiver', 'subsyncer',
                   'rest_api', 'link_handler', 'torrent_handler',
                   'library_handler', 'show_planner', 'torrent_finder',
                   'show_scheduler', 'torrent_cleaner']

    def __init__(self, db=None):
        self._setup_db(db)
        super().__init__(
            'get', self._run, self._omit,
            c_args=(self.releases, self.shows), name='SeriesGetD'
        )
        self.component_map.get('rest_api').foreach(__.set_getd(self))
        self.handlers = self.components.filter_type(Handler)
        self.release_handlers = self.handlers.filter_type(ReleaseHandler)
        self.show_handlers = self.handlers.filter_type(ShowHandler)

    def _setup_db(self, db):
        self.db = db or FileDatabase(self._db_path,
                                     auto_upgrade=self._auto_upgrade_db)
        self.releases = ReleasesFacade(self.db)
        self.shows = ShowsFacade(self.db)

    def load_monitors(self):
        count = self.releases.count
        self.log.info('Loaded {} releases from db.'.format(count))

    def prepare(self):
        self.load_monitors()

    def _cleanup(self):
        self.db.commit()
        self.db.disconnect()

    def activate_release(self, id):
        for c in self.release_handlers:
            c.activate_id(id)

    def activate_show(self, id):
        for c in self.show_handlers:
            c.activate_id(id)

    def explain(self, items: List, handlers: List[Handler],
                services: List[str]):
        handlers = handlers if 'all' in services else (
            services.flat_map(self.component_map.get))
        expl = items / __.explain / handlers.map
        return [dict(item=i.info, expl=e) for i, e in zip(items, expl)]

    def explain_release(self, releases: List[ReleaseMonitor],
                        services: List[str]):
        return self.explain(releases, self.release_handlers, services)

    def format_explain_release(self, releases, services):
        return format_explain_release(self.explain_release(releases, services))

    def explain_show(self, shows: List[Show], services: List[str]):
        return self.explain(shows, self.show_handlers, services)

    def format_explain_show(self, shows, services):
        return format_explain_show(self.explain_show(shows, services))

    @property
    def format_status(self):
        return format_status(self.releases.status)

__all__ = ['SeriesGetD']



# -*- coding: utf-8 -*-

from datetime import datetime

import requests

from golgi.config import configurable
from tek.tools import datetime_to_unix

from series import subsync
from series.get.handler import ReleaseHandler
from series.subsync.errors import NoSubsForEpisode


@configurable(series=['series_dir'], get=['sub_exclude'])
class Subsyncer(ReleaseHandler):

    def __init__(self, releases, *a, **kw):
        super().__init__(releases, 60, 'subtitle downloader')

    def _qualify(self, monitor):
        return (monitor.release.name not in self._sub_exclude and
                monitor.archived and not monitor.subtitles_downloaded and
                monitor.retry_subtitle_download)

    def _handle(self, monitor):
        release = monitor.release
        text = 'Downloading subs for episode {!s}…'
        self.log.info(text.format(release))
        try:
            sub = subsync.get_episode(release.effective_search_name,
                                      release.season,
                                      release.episode)
        except NoSubsForEpisode as e:
            self.log.error('Subsyncer: {}'.format(e))
            f = monitor.subtitle_failures + 1
            self._update(
                monitor, subtitle_failures=f,
                last_subtitle_failure=datetime_to_unix(datetime.now())
            )
        except requests.RequestException as e:
            self.log.error('Subsyncer: {}'.format(e))
        else:
            self._write_sub(monitor, sub)

    def _write_sub(self, monitor, sub):
        try:
            sub.write()
        except IOError as e:
            self.log.error('Subsyncer: {}'.format(e))
        else:
            self._update(monitor, subtitles_downloaded=True)
            self.log.info('Successfully downloaded subtitles.')

__all__ = ['Subsyncer']



from pathlib import Path

from setuptools import setup, find_packages

version_parts = (2, 25, 9)
version = '.'.join(map(str, version_parts))

try:
    from golgi.config.write import write_pkg_config
except ImportError:
    conf = []  # type: ignore
else:
    write_pkg_config('.', Path('series.conf'), 'series')
    conf = ['series.conf']

setup(
    name='series',
    version=version,
    author='Torsten Schmits',
    author_email='torstenschmits@gmail.com',
    license='GPLv3',
    packages=find_packages(exclude=['integration', 'integration.*', 'unit',
                                    'unit.*']),
    data_files=[('share/series/config', conf)],
    package_data={
        '': ['alembic/env.py', 'alembic.ini', 'alembic/versions/*.py'],
    },
    install_requires=[
        'amino>=8.10.4',
        'golgi>=1.2.0',
        'tek-utils>=3.6.2',
        'requests',
        'sqlpharmacy',
        'crystalmethod',
        'lxml',
        'feedparser',
        'flask',
        'alembic',
        'tvrampage',
        'python-dateutil',
        'guessit',
    ],
    entry_points={
        'console_scripts': [
            'serieslibd = series.library.cli:libd',
            'serieslibc = series.library.cli:libc',
            'seriesgetd = series.get.cli:getd',
            'seriesgetc = series.get.cli:getc',
            'subsync_dir = series.subsync:subsync_dir_cli',
            'subsync_cwd = series.subsync:subsync_cwd',
            'subsync_auto = series.subsync:subsync_auto',
            'handle_episode = series.handle_episode:handle_episode_cli',
            'store_episode = series.store_episode:store_episode_cli',
        ]
    }
)
